/**
 * @author
 *  Behrooz Badii 	bb2122@columbia.edu
 *  Hanhua Feng         hanhua@cs.columbia.edu
 *  Edan Harel 	        edan@columbia.edu
 */
package CookieCutter.g6;

import java.util.Vector;
import java.io.*;

/**
 * Solving programs of the Cookie Cutter Problem
 */
public class CookieCutterProblem {
    public static final double DELTA = 1E-5;
    double _sh;         // the height of the strip
    int _nc;            // the number of cookies

    /** the information containing a solution
     */
    public class Solution {
        double[] x;          // x-coordinates of cookies
        double[] y;          // y-coorindates of cookies
        double[] angle;      // angles of cookies
        double rbound = 0;   // the cutting edge

        // the following fields are used to store intermediate data
        int n;               // the number of available solutions
        int[] stage = null;  // the (last) stage that this solution is useful
                             // -1 if the solution is still open
        // int[] idx = null; // the backtrack index
        Polygon[] pgs;       // use spaces to trade-off time
        Rectangle[] bds;         // use spaces to trade-off time
        double bbound = 0;   // the y-dimensional bound
        double bxbound = 0;  

        public Solution( int _n ) {
            x = new double[_n];
            y = new double[_n];
            angle = new double[_n];
            
        }

        public Solution( double[] _x, double[] _y, double[] _angle, int _n ) {
            x = _x;
            y = _y;
            angle = _angle;
            n = _n;
        }

        public Solution copy() {
            Solution ret = new Solution( x.length );
            System.arraycopy( x, 0, ret.x, 0, x.length );
            System.arraycopy( y, 0, ret.y, 0, y.length );
            System.arraycopy( angle, 0, ret.angle, 0, angle.length );
            ret.rbound = rbound;

            return ret;
        }
    }

    /** the constructor
     */
    public CookieCutterProblem( double height, int ncookies ) {
        _sh = height;
        _nc = ncookies;
    }

    /**
     * This solver finds a minimum bounding box and tile it.
     */
    public Solution dumbSolver( Polygon pg ) {

        double rbound = 1E100;
        double angle = 0;
        Rectangle bd = null;
        int n = 0;
        
        // System.out.println( "Bound: " + bound.toString() );
        
        for ( int j=0; j<720; j++ )
        {
            double angle1 = Math.toRadians( j*0.5 );
            Polygon pg1 = pg.rotate( angle1 );
            Rectangle bd1 = pg1.bound();
            
            int n1 = (int) Math.floor( _sh / bd1.height() );
            if ( n1 <= 0 )
                continue;
            
            double rbound1 = ( (_nc-1) / n1 + 1 ) * bd1.width();
            
            if ( rbound1 < rbound )
            {
                n = n1;
                angle = angle1;
                bd = bd1;
                rbound = rbound1;
            }
        }

        if ( null == bd )
            return null;

        Solution ret = new Solution( _nc );
                
        for ( int j=0; j<_nc; j++ )
        {
            ret.y[j] = (j % n) * bd.height() - bd.y0();
            ret.x[j] = (j / n) * bd.width() - bd.x0();
            ret.angle[j] = angle;
        }

        ret.rbound = rbound;
        return ret;
    }

    /**
     * This solver considered the compressable spaces between cookies,
     * It still tile cookies in the same orientation.
     */
    public Solution lessDumbSolver( Polygon pg ) {

        double rbound = 1E100;
        double angle = 0;
        Rectangle bd = null;
        int n = 0;
        double xdist = 0, ydist = 0;
        
        // System.out.println( "Bound: " + bound.toString() );
        
        for ( int j=0; j<180; j+=5 )
        {
            double angle1 = Math.toRadians( j );
            Polygon pg1 = pg.rotate( angle1 );
            Rectangle bd1 = pg1.bound();

            double xdist1 = pg1.contactDistance( pg1 );

            // without this, first upper-level cookie may overlap with
            // third lower-level cookies.
            if ( xdist1 <= bd1.width() * 0.5 )
                xdist1 = bd1.width() * 0.5;

            Polygon[] pgs = new Polygon[] { 
                pg1.translate( xdist1, 0 ), pg1,
                pg1.translate( -xdist1, 0 ) };
            Point p = pg1.contactPos( 
                pgs, 
                new Segment( new Point(0,bd1.height()+DELTA),
                             new Point(0,0) ) );

            double ydist1 = p.distance();
            
            int n1 = (int) Math.floor( (_sh-bd1.height()+ydist1) / ydist1 );
            if ( n1 <= 0 )
                continue;
            
            double rbound1 = ( (_nc-1) / n1 ) * xdist1 + bd1.width();
            
            if ( rbound1 < rbound )
            {
                n = n1;
                angle = angle1;
                bd = bd1;
                rbound = rbound1;
                xdist = xdist1;
                ydist = ydist1;
            }
        }

        if ( null == bd )
            return null;

        // System.out.println( " height(): " + bd.height() + " ydist: " + ydist );
        Solution ret = new Solution( _nc );
                
        for ( int j=0; j<_nc; j++ )
        {
            ret.y[j] = (j % n) * ydist - bd.y0();
            ret.x[j] = (j / n) * xdist - bd.x0();
            ret.angle[j] = angle;
        }

        ret.rbound = rbound;
        return ret;
    }

    /** compute the minimum distances in with different angles
     * n must be even
     */
    static double[][] minDistances1( Polygon pg, int n ) {
        double[][] ret = new double[n][n];
        // int n2 = n/2;
        for ( int i=0; i<n; i++ )
            for ( int j=0; j<n; j++ )
            {
                ret[i][j] =
                    pg.rotate( i*Math.PI*2/n ).contactDistance( 
                        pg.rotate( j*Math.PI*2/n ) );
            }
/*
        for ( int i=n2; i<n; i++ )
        {
            for ( int j=0; j<n2; j++ )
                ret[i][j] = ret[j+n2][i-n2];
            for ( int j=n2; j<n; j++ )
                ret[i][j] = ret[j-n2][i-n2];
        }
*/
        return ret;
    }

    /** get pairwise solution for special matches
     */
    public Solution minDistances2( Polygon pg ) {
        Solution sol = new Solution(8);
        try
        {
            Polygon pg1 = pg.dilate( Plane.EPS );
            // calling experimental functions so we end to hangle exceptions
            int c=0;
            for ( int i=0; i<2; i++ )
            {
                double angle0 = i*Math.PI*0.5;
                for ( int j=0; j<4; j++ )
                {
                    double angle1 = j*Math.PI*0.5;
                    Point p = 
                        perfectMatch( pg1.rotate(angle0), pg1.rotate(angle1) );
                    Point p1 = p.rotate(-angle0);
                    sol.x[c] = p1.x();
                    sol.y[c] = p1.y();
                    sol.angle[c] = angle1 - angle0;
                    c++;
                }
            }
        }
        catch ( RuntimeException e )
        {
            System.out.println( "Experimental code is not good" );
            return null;
        }

        for ( int i=0; i<8; i++ )
        {
            if ( pg.intersects( 
                     pg.place( sol.x[i], sol.y[i], sol.angle[i] ) ) )
            {
                System.out.println( "Exp code have problems ("
                                    + i + "). Ignored it" );
                return null;
            }
        }

        sol.n = 8;
        return sol;
    }

    /** compute the minimum box for pairwise fit, given the minium
     * distance of different angles
     */
    Solution minPair1( Polygon pg, double[][] mindist ) {
        final int m = 24;
        int n = mindist.length;
        Solution ret = new Solution(2);
        double eval = 1E10;
        for ( int i=0; i<n/2; i++ )
        {
            double angle0 = i*Math.PI*2/n;
            Polygon pg0 = pg.rotate( angle0 );
            for ( int j=0; j<n; j++ )
            {
                double angle1 = j*Math.PI*2/n;
                Polygon pg1 = pg.place( mindist[i][j], 0, angle1 );

                for ( int k=0; k<m; k++ )
                {
                    double angle2 = k*Math.PI*2/m;
                    Rectangle bd0 = pg0.rotate( angle2 ).bound();
                    Rectangle bd1 = pg1.rotate( angle2 ).bound();
                    Rectangle bd = bd0.union( bd1 );

                    int c = (int) Math.floor( _sh / bd.height() );
                    if ( c <= 0 )
                        continue;
                    double ev = bd.width() / c;
                    if ( ev < eval )
                    {
                        eval = ev;
                        ret.x[0] = -bd.x0();
                        ret.y[0] = -bd.y0();
                        ret.x[1] = mindist[i][j]*Math.cos( angle2 ) - bd.x0();
                        ret.y[1] = mindist[i][j]*Math.sin( angle2 ) - bd.y0();
                        ret.angle[0] = angle0 + angle2;
                        ret.angle[1] = angle1 + angle2;
                    }
                }
            }
        }
        return ret;
    }

    /** compute the minimum box for pairwise fit, given the pairwise
     * solutions
     */
    Solution minPair2( Polygon pg, Solution mps ) {
        if ( null == mps )
            return null;

        final int m = 24;
        Solution ret = new Solution(2);
        double eval = 1E10;

        for ( int i=0; i<mps.n; i++ )
        {
            double x1 = mps.x[i];
            double y1 = mps.y[i];
            double angle1 = mps.angle[i];
            Polygon pg1 = pg.place( x1, y1, angle1 );

            for ( int k=0; k<m; k++ )
            {
                double angle2 = k*Math.PI*2/m;
                Rectangle bd0 = pg.rotate( angle2 ).bound();
                Rectangle bd1 = pg1.rotate( angle2 ).bound();
                Rectangle bd = bd0.union( bd1 );

                int c = (int) Math.floor( _sh / bd.height() );
                if ( c <= 0 )
                    continue;

                double ev = bd.width() / c;
                if ( ev < eval )
                {
                    eval = ev;
                    Point p = new Point(x1,y1).rotate( angle2 );

                    ret.x[0] = -bd.x0();
                    ret.y[0] = -bd.y0();
                    ret.x[1] = p.x() - bd.x0();
                    ret.y[1] = p.y() - bd.y0();
                    ret.angle[0] = angle2;
                    ret.angle[1] = angle1 + angle2;
                }
            }
        }

        return ret;
    }

    /**
     * dumbSolver, used for pre-arranged cookie pairs (or triples or more)
     */
    public Solution multiDumbSolver( Solution pos, Polygon pg, int ng ) {
        Polygon[] pgs = new Polygon[ng];
        for ( int j=0; j<ng; j++ )
            pgs[j] = pg.place( pos.x[j], pos.y[j], pos.angle[j] );

        double rbound = 1E100;
        double angle = 0;
        Rectangle bd = null;
        int n = 0;
        double xdist = 0, ydist = 0;
        int new_nc = (_nc + ng-1) / ng;
        
        for ( int j=0; j<180; j+=5 )
        {
            double angle1 = Math.toRadians( j );
            Polygon[] pgs1 = new Polygon[ ng ];

            pgs1[0] = pgs[0].rotate( angle1 );
            Rectangle bd1 = pgs1[0].bound();
            for ( int k=1; k<ng; k++ )
            {
                pgs1[k] = pgs[k].rotate( angle1 );
                bd1 = bd1.union( pgs1[k].bound() );
            }

            double xdist1 = 0;
            for ( int k=0; k<ng; k++ )
            {
                double xd = pgs1[k].contactDistance( pgs1 );
                if ( xd >= xdist1 )
                    xdist1 = xd;
            }

            // without this, first upper-level cookie may overlap with
            // third lower-level cookies.
            if ( xdist1 <= bd1.width() * 0.5 )
                xdist1 = bd1.width() * 0.5;

            Polygon[] pgss = new Polygon[ pgs1.length*3 ];
            for ( int k=0; k<pgs1.length; k++ )
            {
                pgss[k+k+k] = pgs1[k].translate( xdist1, 0 );
                pgss[k+k+k+1] = pgs1[k].translate( -xdist1, 0 );
                pgss[k+k+k+2] = pgs1[k];
            }

            double ydist1 = 0;
            Segment seg = new Segment( new Point(0,bd1.height()+DELTA ),
                                       Point.ORIGIN );
            for ( int k=0; k<ng; k++ )
            {
                Point p = pgs1[k].contactPos( pgss, seg );

                double yd = p.distance();
                if ( yd >= ydist1 )
                    ydist1 = yd;
            }
            
            int n1 = (int) Math.floor( (_sh-bd1.height()+ydist1) / ydist1 );
            if ( n1 <= 0 )
                continue;
            
            double rbound1 = ( (new_nc-1) / n1 ) * xdist1 + bd1.width();
            
            if ( rbound1 < rbound )
            {
                n = n1;
                angle = angle1;
                bd = bd1;
                rbound = rbound1;
                xdist = xdist1;
                ydist = ydist1;
            }
        }

        if ( null == bd )
            return null;

        // check whether we can use the dumb solver for the unfilled line
        new_nc = _nc;
        Solution dumb = null;

        int rem = _nc % ( n * ng );
        if ( rem > 0 )
        {
            CookieCutterProblem newprob = new CookieCutterProblem( _sh, rem );

            // System.out.println( "Remaining " +  rem + " n=" + n + " ng=" + ng);

            dumb = newprob.lessDumbSolver( pg );

            // we don't want infinite recursion
            if ( rem < _nc )
            {
                Solution dumb1 = newprob.pairSolver( pg );

                if ( dumb1 != null && dumb1.rbound < dumb.rbound )
                    dumb = dumb1;
            }

            if ( dumb != null && dumb.rbound < xdist )
                new_nc = _nc - rem;
        }
                
        // fill up solutions
        Solution ret = new Solution( _nc );
    outer_loop:
        for ( int i=0, j=0;; i++)
        {
            for ( int k=0; k<ng; k++ )
            {
                Point p = new Point(pos.x[k],pos.y[k]).rotate( angle );

                ret.y[j] = (i % n) * ydist - bd.y0() + p.y();
                ret.x[j] = (i / n) * xdist - bd.x0() + p.x();
                ret.angle[j] = angle + pos.angle[k];
                j++;

                if ( j >= new_nc )
                    break outer_loop;
            }
        }

        if ( new_nc < _nc )
        {
            rbound -= xdist;
            for ( int i=0; i<rem; i++ )
            {
                ret.x[i+new_nc] = dumb.x[i] + rbound;
                ret.y[i+new_nc] = dumb.y[i];
                ret.angle[i+new_nc] = dumb.angle[i];
            }
            rbound += dumb.rbound;
        }

        ret.rbound = rbound;
        return ret;
    }

    /**
     * This solver tries to find a minimum bounding box for a pair of polygons
     */
    public Solution pairSolver( Polygon pg ) {
        Polygon pg_orig = pg;
        pg = pg.dilate( DELTA );

        int nangle = 24;
        Solution mp1 = minPair1( pg, minDistances1( pg, nangle ) );
        Solution mp2 = minPair2( pg, minDistances2( pg ) );
        Solution ret1 = null, ret2 = null;
        double rbound1 = 1E100, rbound2 = 1E100;
        if ( mp1 != null )
        {
            ret1 = multiDumbSolver( mp1, pg, 2 );
            if ( ret1 != null )
                rbound1 = ret1.rbound;
        }

        if ( mp2 != null )
        {
            ret2 = multiDumbSolver( mp2, pg, 2 );
            if ( ret2 != null )
                rbound2 = ret2.rbound;
        }
        
        return ( rbound1 <= rbound2 ? ret1 : ret2 );
    }

    private Solution _kissing_pos = null;
    /** get all available choices that is possible in the next stage
     * @param hist      the locations of all previously placed polygons
     * @param pg        the original polygon, centroid at origin
     * @param prevc     the previous choices
     * @param stage     the current stage
     * @param mindist   the minimum distance at all angles
     * @param yres      the y-dimensional resolution (for the first column)
     */
    public Solution choices( Solution hist, Polygon pg,
                             Solution[] prevc, int stage, 
                             double[][] mindist, double yres ) {

        final int maxc = 1000;

        double[] xc = new double[maxc];
        double[] yc = new double[maxc];
        double[] ac = new double[maxc];
        Polygon[] pgs = new Polygon[maxc];
        Rectangle[] bds = new Rectangle[maxc];
        int[] sc = new int[maxc];
        int nc = 0;
        double radius = pg.radius( Point.ORIGIN );
        double sqrad2 = radius * radius * 4.0;

        // first kind of choices: polygons that attached to the last polygon
        // try all combinations of angles
        if ( stage > 0 )
        {
            // compute the coordinates and angle of the prevoius polygon
            double x0 = hist.x[stage-1];   
            double y0 = hist.y[stage-1];
            double angle0 = hist.angle[stage-1];

        loop1:
            for ( int i=0; i<mindist.length; i++ ) 
            {
                // the actual direction of the new polygon is the
                // difference of angle the previous polygon and its
                // anticipated angle w.r.t. the new polygon.
                double angle1 = i*2*Math.PI/mindist.length;
                double cos = Math.cos( angle0-angle1 );
                double sin = Math.sin( angle0-angle1 );

            loop2:
                for ( int j=0; j<mindist.length; j++ )
                {
                    double angle2 = j*2*Math.PI/mindist.length;
                    double dist = mindist[i][j];
                    double x = dist * cos + x0;
                    double y = dist * sin + y0;
                    double angle = angle0-angle1+angle2; // the actual angle 
/*
                    System.out.println( "Trying all combination of angles:"
                                        + Math.toDegrees(angle1) + " vs " 
                                        + Math.toDegrees(angle2) );
*/
                    if ( x <= 0 || y <= 0 || y >= _sh )
                        continue;

                    Polygon pgtmp = pg.place( x, y, angle );
                    Rectangle bd = pgtmp.bound();

                    if ( bd.x0() < 0 || bd.y0() < 0 || bd.y1() >= _sh )
                        continue;

                    for ( int k=0; k<stage-1; k++ )
                    {
                        if ( Point.sqdistance( x, y, hist.x[k], hist.y[k] ) 
                             <= sqrad2 && pgtmp.intersects( hist.pgs[k] ) )
                            continue loop2;
                    }
                    
                    xc[nc] = x;
                    yc[nc] = y;
                    ac[nc] = angle;
                    sc[nc] = -1;
                    pgs[nc] = pgtmp;
                    bds[nc] = bd;
/*
                    System.out.println( 
                        "Stage #" + stage + ": Add a first-kind choice #"
                        + nc + " (" + xc[nc] + "," + yc[nc] + ";" 
                        + Math.toDegrees(ac[nc]) + ")" );
*/
                    if ( ++nc >= maxc )
                        break loop1;
                }
            }
        }

        // set up the new bxbound. (bottom extension bound)
        double bxbound = pg.radius( new Point(0,0) ) * 2;
        if ( stage > 0 )
            bxbound += prevc[stage-1].bbound;

        // the second kind of choices: position available at the
        // left-most column
        if ( nc < maxc )
        {
            double oldbxb = 0;
            if ( stage > 0 )
                oldbxb = prevc[stage-1].bxbound;
            // System.out.println( "Oldbxb: " + oldbxb );

        loop4:
            for ( int j=0; j<mindist.length; j++ )
            {
                double angle = j*2*Math.PI/mindist.length;
                Polygon pgtmp0 = pg.rotate( angle );
                Rectangle bd0 = pgtmp0.bound();
                double ymax = Math.min( bxbound, _sh - bd0.height() );

            loop5:
                for ( double yoff=oldbxb; yoff<=ymax; yoff += yres ) 
                {
                    double x = -bd0.x0();
                    double y = yoff-bd0.y0();
                    Polygon pgtmp = pgtmp0.translate( x, y );
                    Rectangle bd = new Rectangle( 0, yoff, 
                                                  bd0.width(), bd0.height() );

                    for ( int k=0; k<stage; k++ )
                    {
                        if ( Point.sqdistance( x, y, hist.x[k], hist.y[k] ) 
                             <= sqrad2 && pgtmp.intersects( hist.pgs[k] ) )
                            continue loop5;
                    }

                    xc[nc] = x;
                    yc[nc] = y;
                    ac[nc] = angle;
                    sc[nc] = -1;
                    pgs[nc] = pgtmp;
                    bds[nc] = bd;
/*
                    System.out.println( 
                        "Stage #" + stage + ": Add a second-kind choice #"
                        + nc + " (" + xc[nc] + "," + yc[nc] + ";" 
                        + Math.toDegrees(ac[nc]) + ")" );
*/
                    if ( ++ nc >= maxc )
                        break loop4;
                }
            }
        }

        // the third kind of choices
        if ( nc < maxc && _kissing_pos != null && stage > 0 )
        {
            double x0 = hist.x[stage-1];   
            double y0 = hist.y[stage-1];
            double angle0 = hist.angle[stage-1];
        loop7:
            for ( int i=0; i<_kissing_pos.n; i++ )
            {
                Point p = new Point( _kissing_pos.x[i], _kissing_pos.y[i] )
                    .rotate( angle0 );
                
                double x = x0 + p.x();
                double y = y0 + p.y();
                double angle = _kissing_pos.angle[i] + angle0;
                Polygon pgtmp = pg.place( x, y, angle );
                Rectangle bd = pgtmp.bound();

                if ( bd.x0() < 0 || bd.y0() < 0 || bd.y1() >= _sh )
                    continue;

                for ( int k=0; k<stage; k++ )
                {
                    if ( Point.sqdistance( x, y, hist.x[k], hist.y[k] ) 
                         <= sqrad2 && pgtmp.intersects( hist.pgs[k] ) )
                        continue loop7;
                }
                
                xc[nc] = x;
                yc[nc] = y;
                ac[nc] = angle;
                sc[nc] = -1;
                pgs[nc] = pgtmp;
                bds[nc] = bd;
/*
                System.out.println( 
                    "Stage #" + stage + ": Add a third-kind choice #"
                    + nc + " (" + xc[nc] + "," + yc[nc] + ";" 
                        + Math.toDegrees(ac[nc]) + ")" );
*/
                if ( ++ nc >= maxc )
                    break;
            }
        }

        prevc[stage] = new Solution( xc, yc, ac, nc );
        prevc[stage].stage = sc;
        prevc[stage].pgs = pgs;
        prevc[stage].bds = bds;
        prevc[stage].bxbound = bxbound;

        // the forth kind of choices: valid previous choices
        if ( nc < maxc )
        {
            // previous choices
        loop3:
            for ( int i=stage-1; i>=0; i-- )
                for ( int j=0; j<prevc[i].n; j++ )
                    if ( prevc[i].stage[j] < 0 ) 
                    {
                        xc[nc] = prevc[i].x[j];
                        yc[nc] = prevc[i].y[j];
                        ac[nc] = prevc[i].angle[j];
                        sc[nc] = i;
                        pgs[nc] = prevc[i].pgs[j];
                        bds[nc] = prevc[i].bds[j];
/*
                        System.out.println( 
                            "Stage #" + stage + ": Add a forth-kind choice #"
                            + nc + " (" + xc[nc] + "," + yc[nc] + ";" 
                            + Math.toDegrees(ac[nc]) + ")" );
*/
                        if ( ++nc >= maxc )
                            break loop3;
                    }
        }

        Solution ret = new Solution( xc, yc, ac, nc );
        ret.stage = sc;
        ret.pgs = pgs;
        ret.bds = bds;

        return ret;
    }

    /** choose the idx-th choices and update the database
     */
    public void choose( Solution choices, int idx, 
                        Solution hist, Polygon pg,
                        Solution[] prevc, int stage ) {

        double x = choices.x[idx];
        double y = choices.y[idx];
        double angle = choices.angle[idx];
        Polygon npg = pg.place( x, y, angle );
        double radius = pg.radius( Point.ORIGIN ) + Plane.EPS;
        double sqrad2 = radius * radius * 4.0;
/*
        System.out.println( "Stage #" + stage + ": choose choice #" + idx
                            + " of " + choices.n + ": (" + x + "," + y + ";" 
                            + Math.toDegrees(angle) + ")" );
*/
        Rectangle r = npg.bound();

        hist.x[stage] = x;
        hist.y[stage] = y;
        hist.angle[stage] = angle;
        hist.pgs[stage] = npg;
        hist.bds[stage] = r;

        if ( r.x1() > hist.rbound )
            hist.rbound = r.x1();
        if ( r.y1() > hist.bbound )
            hist.bbound = r.y1();

        prevc[stage].rbound = hist.rbound;
        prevc[stage].bbound = hist.bbound;

        // eliminate now-invalid previous choices.
        for ( int i=0; i<=stage; i++ )
        {
            Solution prevci = prevc[i];
            for ( int j=0; j<prevc[i].n; j++ )
                if ( prevci.stage[j] < 0 && Point.sqdistance( 
                         x, y, prevci.x[j], prevci.y[j] ) <= sqrad2
                     && npg.intersects( prevci.pgs[j] ) )
                {
                    prevc[i].stage[j] = stage;
/*
                    System.out.println( 
                        "Stage #" + stage + ": remove choice #" + j
                        + " of " + prevc[i].n + " from stage #" + i
                        + " (" + prevc[i].x[j] + "," + prevc[i].y[j] + ";" 
                        + Math.toDegrees(prevc[i].angle[j]) + ")" );
*/
                }
        }
    }

    /** for backtrack algorithm
     */
    public void backtrack( Solution[] prevc, int stage ) {
        for ( int i=0; i<stage; i++ )
            for ( int j=0; j<prevc[i].n; j++ )
                if ( prevc[i].stage[j] == stage )
                    prevc[i].stage[j] = -1;
    }

    /** An greedy algorithm with random initial placement
     */
    public Solution greedySolver( Polygon pg ) {
        // System.out.println( "Initiate greedy algorithm" );
        _kissing_pos = minDistances2( pg );

        final int nangle = 12;
        double[][] mindist = minDistances1( pg, nangle );

        Solution ret = new Solution( _nc );
        // System.out.println( "running greedy algorithm" );

        ret.rbound = 1E100;
        for ( int k=0; k<2; k++ )
        {
            Solution soln = new Solution( _nc );
            soln.pgs = new Polygon[ _nc ];
            soln.bds = new Rectangle[ _nc ];
            Solution[] prevc = new Solution[ _nc ];

            for ( int i=0; i<_nc; i++ )
            {
                Solution choices = choices( soln, pg, prevc, i, 
                                            mindist, _sh*0.1 );

                if ( 0 == choices.n )
                    break;

                int idx = -1;
                double rbound = 1E100;

                double ratio = (double)(_nc-i-1) / (double)_nc;
                
                for ( int j=0; j<choices.n; j++ )
                {
                    double rb = choices.bds[j].x1() 
                        + 0.9 * choices.bds[j].y1() * ratio;
                    if ( rb < rbound ) 
                    {
                        rbound = rb;
                        idx = j;
                    }
                }

                if ( k > 0 && i == 0 )
                    idx = Plane.rand.nextInt( choices.n );

                choose( choices, idx, soln, pg, prevc, i );
            }            

            if ( soln.rbound < ret.rbound )
                ret = soln;
        }

        if ( ret.rbound >= 1E50 )
            return null;

        return ret;
    }

    /** try to compress existing solution and generate a new solution
     */
    public Solution compress( Solution soln, Polygon pg ) {
        Solution ret = soln.copy();

        final int n2dir = 6;
        double[] alpha = new double[ n2dir * 2 + 1 ];
        Segment[] dirs = new Segment[ n2dir * 2 + 1 ];
       
        // create testing directions
        alpha[0] = Math.PI;
        dirs[0] = new Segment( Point.ORIGIN, new Point( -10, 0 ) );
        for ( int i=1; i<=n2dir; i++ )
        {
            alpha[i+i-1] = Math.PI-i*Math.PI*0.5/n2dir;
            double cos = Math.cos(alpha[i+i-1]);
            double sin = Math.sin(alpha[i+i-1]);
            dirs[i+i-1] = new Segment( Point.ORIGIN,
                                       new Point( 10*_sh*cos, 10*_sh*sin ) );
            alpha[i+i] = -alpha[i+i-1];
            dirs[i+i] = new Segment( Point.ORIGIN,
                                     new Point( 10*cos, -10*sin ) );
        }

        ret.pgs = new Polygon[_nc];
        for ( int i=0; i<_nc; i++ )
            ret.pgs[i] = pg.place( ret.x[i], ret.y[i], ret.angle[i] );

        Polygon box = new Polygon( new Rectangle( 0, 0, _nc*10, _sh ) );
        Polygon[] rest = new Polygon[_nc];
        for ( int i=0; i<_nc; i++ )
            rest[i] = ret.pgs[i];

        for ( int k=0; k<3; k++ )
        {
            for ( int i=0; i<_nc; i++ )
            {
                rest[i] = box;
                Polygon pg0 = ret.pgs[i];

                double eval = 0;
                double x = 0;
                double y = 0;
                
                for ( int j=0; j<dirs.length; j++ )
                {
                    Point p = pg0.contactPos( rest, dirs[j] );
                    double ev = p.x()*p.x() + 0.01 * p.y() * p.y();
                    if ( ev > eval )
                    {
                        eval = ev;
                        x = p.x();
                        y = p.y();
                    }
                }

                double d = Math.sqrt(x*x+y*y);
                if ( d > DELTA * 5.0 )
                {
                    x -= DELTA * x / d;
                    y -= DELTA * y / d;

                    ret.x[i] += x;
                    ret.y[i] += y;
                    ret.pgs[i] = pg0.translate( x, y );
                }

                rest[i] = ret.pgs[i];
            }
        }

        ret.rbound = calcRBound( ret.pgs );
        return ret;
    }

    /**
     * calculate the right bounds of polygons
     */
    public double calcRBound( Polygon[] pgs ) {
        double ret = 0;
        for ( int i=0; i<pgs.length; i++ )
        {
            double d = pgs[i].bound().x1();
            if ( d > ret )
                ret = d;
        }

        return ret;
    }

/*
 * TEST TEST TEST TEST TEST TEST TEST TEST TEST TEST TEST TEST
 * The test program
 * TEST TEST TEST TEST TEST TEST TEST TEST TEST TEST TEST TEST
 */
    public boolean valid( Polygon pg, CookieCutterProblem.Solution soln ) {

        for ( int i=0; i<_nc; i++ )
        {
            Polygon pg0 = pg.place( soln.x[i], soln.y[i], soln.angle[i] );
            
            Rectangle r = pg0.bound();
                
            if ( r.x0() <= 0 || r.x1() >= soln.rbound
                 || r.y0() <= 0 || r.y1() >= _sh )
            {
                System.out.println( "Warning: cookie #" + i 
                                    + " is out of the bounding box." );
                return false;
            }
                
            for ( int j=0; j<i; j++ )
            {
                Polygon pg1 = pg.place( soln.x[j], soln.y[j], soln.angle[j] );
                if ( pg0.intersects( pg1 ) )
                {
                    System.out.println( "Warning: cookie #" + i 
                                        + " intersects with " + j );
                    return false;
                }                
            }
        }

        return true;
    }

    public static void main( String[] args ) {
        Painter pane = Painter.create( "The Lord of Cookie Cutters", 980, 830 );
        Painter.main = pane;

        double xdist = 4.8;
        double ydist = 1.05;
        double nsol = 4;

        pane.setRange( -0.05, xdist+0.05, -0.05, ydist*4+0.05 );

        Polygon[] pgs = null;

        if ( args.length >= 1 ) 
        {
            pgs = Painter.readGameFile( args[0] );

            if ( null == pgs )
            {
                System.err.println( "File not found: " + args[0] );
                return;
            }
        }

        int n = 10;
        if ( args.length >= 2 )
            n = Integer.parseInt( args[1] );

        int[] base_color = new int[]{ 0xff0000, 0x00ff00, 0x808000, 0xffff00 };

        for ( int i=0; pgs != null && i<pgs.length; i++ )
        {
            Point centroid = pgs[i].centroid();
            Polygon pg = pgs[i].translate( -centroid.x(), -centroid.y() );

            if ( n < 0 )
            {
                for ( int j=0; j<4; j++ )
                {
                    double x = j%2*1.2+0.5;
                    double y = j/2*1.2+2;
                    Painter.main.draw( pg.place( x, y, Math.PI*0.5*j ), 
                                       0xb0b0b0 );
                    Painter.main.draw( pg.translate( x, y ), 0 );
                    
                    Point p = perfectMatch( pg,
                                            pg.rotate( Math.PI*0.5*j ) );
                    if ( p != null )
                    {
                        // System.out.println( "Matching: " + p.toString() );
                        Painter.main.draw( 
                            pg.place( x+p.x(), y+p.y(), 
                                      Math.PI*0.5*j ), 0xffff00 );
                    }
                }


                pane.waitClick();
                pane.clear();
                continue;
            }


            Polygon[] hist = new Polygon[n];

            CookieCutterProblem prob = new CookieCutterProblem( 1.0, n );
            Polygon pg0 = pg.dilate(DELTA);

            Solution soln[] = {
                prob.lessDumbSolver( pg0.dilate(DELTA) ),
                prob.pairSolver( pg0.dilate(DELTA) ) ,
                prob.greedySolver( pg0.dilate(DELTA) ),
                prob.compress( prob.pairSolver( pg0.dilate(DELTA) ), pg0 )
            };

            for ( int k=0; k<nsol; k++ )
            {
                pane.draw( new Rectangle( 0, ydist*k, xdist, 1 ), 0x808080 );

                if ( null == soln[k] )
                    pane.draw( "Solution " + k + " does not exist", 
                               base_color[k] );
                else if ( !prob.valid( pg, soln[k] ) )
                    pane.draw( "Solution " + k + " is invalid", 
                               base_color[k] );
            }

            for ( int j=0; j<n; j++ )
            {
                for ( int k=0; k<nsol; k++ )
                {
                    if ( soln[k] != null )
                    {
                        double yoff = ydist * (nsol-k-1);
                        Polygon rpg = pg.place( soln[k].x[j], soln[k].y[j], 
                                                soln[k].angle[j] );
                        pane.draw( rpg.translate( 0, yoff ), 
                                   0xc0000000+base_color[k] + (j*95+n)%256 );

                        pane.draw( new Segment( 
                                       new Point( soln[k].rbound, yoff ),
                                       new Point( soln[k].rbound, yoff+1 ) ), 
                                   0 );
                    }
                }
            }

            pane.waitClick();
            pane.clear();
        }

        int width = (int) Math.floor( xdist );
        if ( pgs != null ) 
        {
            for ( int i=0; i<pgs.length; i++ )
                pane.draw( pgs[i].translate( i%width, i/width*2.0 ), 0x1000000 );
            // i/pgs.length );
        }

        Polygon E = new Polygon( new Point[] {
            new Point(0,0), new Point(0,1),  new Point(0.8,1), 
            new Point(0.8,0.8), new Point(0.2,0.8), new Point(0.2,0.6),
            new Point(0.6,0.6), new Point(0.6,0.4), new Point(0.2,0.4), 
            new Point(0.2,0.2), new Point(0.8,0.2), new Point(0.8,0) } );
        Polygon N = new Polygon( new Point[] {
            new Point(0,0), new Point(0,1), new Point(0.3,1), 
            new Point(0.6,0.2), new Point(0.6,1), new Point(0.8,1),
            new Point(0.8,0), new Point(0.5,0),
            new Point(0.2,0.8), new Point(0.2,0) } );
        Polygon D1 = new Polygon( new Point[] {
            new Point(0,0), new Point(0,1), new Point(0.2,1), 
            new Point(0.2,0) } );
        Polygon D2 = new Polygon( new Point[] {
            new Point(0.2,1), new Point(0.4,0.95), new Point(0.7,0.8), 
            new Point(0.8,0.5), new Point(0.7,0.2), new Point(0.4,0.05),
            new Point(0.2,0), new Point(0.2,0.2), new Point(0.5,0.3), 
            new Point(0.6,0.5), new Point(0.5,0.7), new Point(0.2,0.8 ) } );

        pane.draw( E.translate( 1, 1 ), 0xffff0000 );
        pane.draw( N.translate( 2, 1 ), 0xff00ff00 );
        pane.draw( D1.translate( 3, 1 ), 0xff0000ff );
        pane.draw( D2.translate( 3, 1 ), 0xff0000ff );
    }


/*
 * EXPERIMENTAL EXPERIMENTAL EXPERIMENTAL EXPERIMENTAL EXPERIMENTAL
 * The following functions are experimental.
 * EXPERIMENTAL EXPERIMENTAL EXPERIMENTAL EXPERIMENTAL EXPERIMENTAL
 */

    private static double[] __interval_and( double[] iv0, int n0, 
                                            double[] iv1, int n1 ) {
        double[] iv = new double[ n0+n1 ];
        int niv = 0;

    outer_loop:
        for ( int i=0, j=0;;)
        {
            while ( iv0[i+1] <= iv1[j] )
            {
                i += 2;
                if ( i >= n0-1 )
                    break outer_loop;
            }

            while ( iv0[i] >= iv1[j+1] )
            {
                j += 2;
                if ( j >= n1-1 )
                    break outer_loop;
            }

            iv[niv++] = Math.max( iv0[i], iv1[j] );
            iv[niv++] = Math.min( iv0[i+1], iv1[j+1] );


            if ( iv0[i+1] >= iv1[j+1] )
            {
                j += 2;
                if ( j >= n1-1 )
                    break;
            }
            else
            {
                i += 2;
                if ( i >= n0-1 )
                    break;
            }
        }

        if ( niv == 0 )
            return null;

        if ( niv == iv.length )
            return iv;
        double[] tmp = new double[niv];
        System.arraycopy( iv, 0, tmp, 0, niv );
        return tmp;
    }

    private static double __interval_area( double[] data, int ndata ) {
        double ret = 0;
        for ( int i=0; i<ndata-1; i++ )
            ret += data[i+1]-data[i];
        return ret;
    }


    private static double[] _compute_interval( 
        Polygon pg0, Polygon pg1, double xinf, double xsup, double y ) {

        double[] itv = new double[pg0.n()+2];
        double[] ret = new double[] { xinf, xsup };

        for ( int i=0; i<pg1.n(); i++ )
        {
            Segment seg = new Segment(
                new Point( pg1.x(i)+xinf, pg1.y(i)+y ),
                new Point( pg1.x(i)+xsup, pg1.y(i)+y ) );

            int nitv = 0;

            itv[nitv++] = xinf - pg1.x(i);
            for ( int j=0; j<pg0.n(); j++ )
            {
                Object o = seg.intersection( pg0.edge(j) );
                if ( null == o || ! ( o instanceof Point ) )
                    continue;
                Point p = (Point) o;
                
                itv[nitv++] = p.x() - pg1.x(i);
            }
            itv[nitv++] = xsup - pg1.x(i);

            if ( ( nitv & 1 ) == 1 )
            {
                System.out.println( "odd number" );
                continue;
            }

            java.util.Arrays.sort( itv, 0, nitv );

            ret = __interval_and( ret, ret.length, itv, nitv );

            if ( ret == null )
                return null;
/*
            if ( ___k == 64/2+6 )
            {

            System.out.println( "Itv for y = " + y + " i=" + i );
            for ( int t=0; t<nitv; t++ )
                System.out.print( " " + Plane.format(itv[t])  );
            System.out.println();
            
            if ( ret != null )
            {
                System.out.println( "interval for y = " + y + " i=" + i);
                for ( int t=0; t<ret.length; t++ )
                    System.out.print( " " + Plane.format(ret[t])  );
                System.out.println();
            }

            double yp = pg1.y(i)+y;
            for ( int j=0; j<nitv; j+=2 )
            {                            
                Painter.main.draw( 
                    new Segment( 
                        new Point( itv[j]+1+pg1.x(i), yp+2.5 ),
                        new Point( itv[j+1]+1+pg1.x(i), yp+2.5 )),
                    0xff00 );
                
                Painter.main.draw( 
                    new Segment( 
                        new Point( itv[j]+3, yp-yp+2.5+i*0.01 ),
                        new Point( itv[j+1]+3, yp-yp+2.5+i*0.01 )),
                    0xff00 );
            }
            
                if ( null != ret )
                    for ( int j=0; j<ret.length; j+=2 )
                    {
                        Painter.main.draw( 
                            new Segment( 
                                new Point( ret[j]+3, 2.1-i*0.1 ),
                                new Point( ret[j+1]+3, 2.1-i*0.1 )),
                            0xff0000 );
                    }
                
            }
*/
        }

        return ret;
    }

    private static double __smallest_abs( Polygon pg0, Polygon pg1, 
                                          double[] itv, double y ) {
        double ret = 1E100;
        for ( int i=0; i<itv.length; i++ )
        {
            if ( Math.abs( itv[i] ) < Math.abs( ret ) )
            {
                double x = ( (i&1)==1 ? itv[i]-DELTA : itv[i]+DELTA );
                if ( !pg1.translate( x, y ).intersects( pg0 ) )
                    ret = itv[i];
            }
        }

        return ret;
    }

    public static Point perfectMatch( Polygon pg0, Polygon pg1 ) {
        Rectangle bd0 = pg0.bound();
        Rectangle bd1 = pg1.bound();
        double ysup = bd0.y1() - bd1.y0() - DELTA;
        double yinf = bd0.y0() - bd1.y1() + DELTA;
        double xsup = ( bd0.x1() - bd1.x0() ) * 2.01 + DELTA;
        double xinf = ( bd0.x0() - bd1.x1() ) * 2.01 - DELTA;
        Point ret = null;
        double d = 1E100;

        final int np = 128;
        for ( int k=0; k<np; k++ ) 
        {
            double x, y = (k+0.5)*(ysup-yinf)/np + yinf;

            // ___k = 0;
            double[] itv1 =
                _compute_interval( pg0, pg1, xinf, xsup, y );
            if ( itv1 == null )
                continue;
            
            // ___k = k;

            double[] itv2 = 
                _compute_interval( pg1, pg0, -xsup, -xinf, -y );
            if ( itv2 == null )
                continue;
            
            for ( int i=itv2.length/2-1; i>=0; i-- )
            {
                double tmp = itv2[i];
                itv2[i] = -itv2[itv2.length-i-1];
                itv2[itv2.length-i-1] = -tmp;
            }
            
            double[] itv = 
                __interval_and( itv1, itv1.length, itv2, itv2.length );
/*
            if ( ___k == 64/2+6 )
            {

                for ( int t=0; t<itv1.length; t++ )
                    System.out.print( " " + Plane.format(itv1[t])  );
                System.out.println();


                for ( int t=0; t<itv2.length; t++ )
                    System.out.print( " " + Plane.format(itv2[t])  );
                System.out.println();

                for ( int t=0; t<itv.length; t++ )
                    System.out.print( " " + Plane.format(itv[t])  );
                System.out.println();
            }
*/

            if ( itv != null )
            { 
                x = __smallest_abs( pg0, pg1, itv, y );
/*
                if ( k == np/2+6 ) {
                    System.out.println( "Smallest x:" + x + " contactdist: " + __contact_dist( pg0, pg1, p ) + " cont" + __contact_dist( pg1, pg1, p ) );
                    
                    Painter.main.draw( pg1.translate(x+1,y+2.5), 0xff0000 );
                }
*/                
                double dist = Math.sqrt( x*x+y*y );
                if ( dist < d )
                {
                    d = dist;
                    ret = new Point(x,y);
                }
            }
        }
        
        return ret;
    }

    public static final double DELTA2 = 1E-3;
    /** remove polygon pg from a simple region
     */
    public static SimpleRegion[] difference( 
        SimpleRegion reg,  Polygon pg ) throws GeometryException {

        // add all crosspoint to polygons reg.pg and pg
        Polygon[] pgs = reg.pg.crossify( pg );

        // if there is no change for both polygons
        if ( null == pgs )
        {
            // check the relationship between two original polygons
            int cmp = pg.compare( reg.pg );
            // if pg is a superset of the polygon, return null;
            if ( cmp == Polygon.CMP_SUPER )
                return null;

            // if the root polygon of the region covers pg, add
            // pg to the holes
            if ( cmp == Polygon.CMP_SUB )
            {
                Polygon[] nh;
                if ( null == reg.holes || reg.holes.length == 0 )
                    nh = new Polygon[] { pg };
                else
                {
                    nh = new Polygon[reg.holes.length+1];
                    System.arraycopy( reg.holes, 0, nh, 1, reg.holes.length );
                    nh[0] = pg;
                }
                
                return new SimpleRegion[] { new SimpleRegion ( reg.pg, nh ) };
            }

            // otherwise, the region is not affected
            return new SimpleRegion[] { reg };
        }

        // compute the cut sets of difference
        Polygon[][] cutsets = pgs[0].cutsets( pgs[1] );
        if ( cutsets == null )
            return null;
        Polygon[] cut = cutsets[1];
        if ( null == cut || cut.length == 0 )
            return null;

        // System.out.println( "Cut length: " + cut.length );

        Vector plist = new Vector();
        // for all cut sets
        for ( int i=0; i<cut.length; i++ )
        {
            // create a region contains only one of the cut polygon
            SimpleRegion cutreg = new SimpleRegion( cut[i], null );

            // if there is no holes in the original polygon, add
            // this region to the result
            if ( reg.holes == null || reg.holes.length == 0 )
            {
                plist.add( cutreg );
                continue;
            }
/*
            System.out.println( "Checking " + reg.holes.length 
                                + " holes: " );
*/
            // if there are some holes, we have to do something
            // more complicated. 
            SimpleRegion[] tmp = new SimpleRegion[] { cutreg };

            // try to remove holes from the cut polygon
            for ( int j=0; j<reg.holes.length; j++ )
            {
                tmp = difference( tmp, reg.holes[j] );
                if ( tmp == null )
                    break;
            }

            // if some regions of the cut polygon are not covered by
            // the holes, add then to the result
            if ( tmp != null )
            {
                for ( int k=0; k<tmp.length; k++ )
                    plist.add( tmp[k] );
            }
        }

        return (SimpleRegion[]) plist.toArray( new SimpleRegion[0] );
    }

    /** remove polygon to a region set
     */
    public static SimpleRegion[] difference( 
        SimpleRegion[] reg, Polygon pg ) throws GeometryException {

        Vector plist = new Vector();
        // for each region, remove pg from it and add the returned
        // region set to the result
        for ( int i=0; i<reg.length; i++ )
        {
/*
            System.out.println( "Computing the difference of " 
                                + reg[i].toString() + " and " 
                                + pg.toString() );
*/
            SimpleRegion[] tmp = null;

            Polygon pg1 = pg;
            for ( int k=0; ; k++ )
            {
                try
                {
                    tmp = difference( reg[i], pg );
                }
                catch ( GeometryException e )
                {
                    // e.printStackTrace();
                    // throw new RuntimeException();
                    // System.out.println( "Error detected" );

                    pg1 = pg1.dilate( DELTA2*k*k );
                    if ( k > 10 )
                        throw e;
                    continue;

                }
                break;
            }
/*            
            System.out.println( "The difference is "
                                + SimpleRegion.toString( tmp ) );
*/
            if ( tmp != null )
            {
                for ( int k=0; k<tmp.length; k++ )
                    plist.add( tmp[k] );
            }
        }
        
        return ( SimpleRegion[] ) plist.toArray( new SimpleRegion[0] );
    }

    public static Polygon crossingOffset( Segment s0, Segment s1 ) {
        double x0 = ( s0.x0() + s0.x1() ) * 0.5;
        double y0 = ( s0.y0() + s0.y1() ) * 0.5;
        double x1 = ( s1.x0() + s1.x1() ) * 0.5;
        double y1 = ( s1.y0() + s1.y1() ) * 0.5;
        double x = x1 - x0;
        double y = y1 - y0;
        double dx0 = ( s0.x1() - s0.x0() ) * 0.5;
        double dy0 = ( s0.y1() - s0.y0() ) * 0.5;
        double dx1 = ( s1.x1() - s1.x0() ) * 0.5;
        double dy1 = ( s1.y1() - s1.y0() ) * 0.5;

        Polygon ret = new Polygon ( new Point[] {
            new Point( x + dx0 + dx1, y + dy0 + dy1 ), 
            new Point( x + dx0 - dx1, y + dy0 - dy1 ),
            new Point( x - dx0 - dx1, y - dy0 - dy1 ), 
            new Point( x - dx0 + dx1, y - dy0 + dy1 ) } );

        return ret.abs().dilate( DELTA2 + DELTA2 * Math.random() );
    }

    public static SimpleRegion[] offsetRegions( SimpleRegion[] prev,
                                                Polygon last, Polygon pg ) {
        
        SimpleRegion[] ret = prev;
        for ( int i=0; i<last.n(); i++ )
            for ( int j=0; j<pg.n(); j++ )
            {
                Polygon pg1 = crossingOffset( pg.edge(j), last.edge(i) );
                Rectangle bd1 = pg1.bound();
                boolean flag = true;
                for ( int k=0; k<ret.length; k++ )
                    if ( bd1.intersects( ret[k].bound ) )
                    {
                        flag = false;
                        break;
                    }

                if ( flag )
                    continue;
                
                for ( int k=0; k<10; k++ )
                {
                    try 
                    {
                        ret = difference( ret, pg1 );
                    } 
                    catch ( GeometryException e)
                    {
                        pg1 = pg1.dilate( DELTA2*k );
                        continue;
                    }
                }
/*
                if ( Painter.main != null )
                {
                    Painter.main.draw( pg1, counter*20 );
                    if ( ret != null && ret.length > 0 )
                    {
                        for ( int k=0; k<ret.length; k++ )
                        {
                            Painter.main.draw( ret[k].pg, 
                                           0xffffff - counter*0x3475 );
                        }
                        System.out.println( "===========================" );
                        System.out.println( "Regions: " + SimpleRegion.toString( ret ) );
                        System.out.println( "Number of regions: " + 
                                            ret.length );
                                            
                    }
                    else
                    {
                        System.out.println( "No regions available" );
                        Painter.waitClick();
                    }

                    counter++;
                    // Painter.waitClick();
                }
*/
            }
        return ret;
    }

    public Solution greacySolver( Polygon pg ) {
        final int nr_angles = 12;
        Polygon[] rpgs = new Polygon[nr_angles];
        SimpleRegion[][] regsets = new SimpleRegion[nr_angles][];
        Line ln = new Line( new Point(0,0), new Point(-0.0001, 1) );
        Polygon[] hist = new Polygon[_nc];
        Solution ret = new Solution(_nc);

        Rectangle rect = new Rectangle( 0, 0, 300, _sh );

        for ( int i=0; i<nr_angles; i++ )
        {
            rpgs[i] = pg.rotate( i*Math.PI*2/nr_angles );
            Rectangle bd = rpgs[i].bound();
            regsets[i] = new SimpleRegion[] { 
                new SimpleRegion(
                    new Polygon(  
                        new Rectangle( 
                            rect.x0() - bd.x0(), rect.y0() - bd.y0(), 
                            rect.width() - bd.width(),
                            rect.height() - bd.height() ) ),
                    null ) };
            
        }

        double rbound = 0;
        for ( int i=0; i<_nc; i++ )
        {
            double dist = 1E100;
            Point pos = null;
            double angle = 0;
            for ( int j=0; j<nr_angles; j++ )
            {
                Point p = closestVertex( regsets[j], ln );
                if ( p == null )
                    continue;
                if ( _min_dist < dist )
                {
                    dist = _min_dist;
                    pos = p;
                    angle = j*Math.PI*2/nr_angles;
                }
            }

            // System.out.println( "Cookie #" + i );
            ret.x[i] = pos.x();
            ret.y[i] = pos.y();
            ret.angle[i] = angle;
            hist[i] = pg.rotate(angle).translate( ret.x[i], ret.y[i] );
            Rectangle bd1 = hist[i].bound();
            if ( bd1.x1() > ret.rbound ) 
                ret.rbound = bd1.x1();

            for ( int j=0; j<nr_angles; j++ )
                regsets[j] = offsetRegions( regsets[j], hist[i], rpgs[j] );
        }
        return ret;
    }

    static double _min_dist = 1E100;
    public static Point closestVertex( SimpleRegion[] regs, Line ln ) {
        if ( regs == null || regs.length == 0 )
            return null;

        Point p = null;
        double x = 1E100;

        for ( int i=0; i<regs.length; i++ )
        {
            Point px = regs[i].pg.closest( ln );
            double dist = ln.distance( px );
            if ( dist < x )
            {
                p = px;
                x = dist;
            }
        }

        _min_dist = x;
        
        return p;
    }

    public Solution PuzzleSolver( Polygon shape1, Polygon shape2) {
        double currentminx = -1;
        int angle;
        int testnum=1;
        int extrapiece=0;
        double xtomove;
        double ytomove;
        double x2tomove;
        double y2tomove;
        double x2finalmove=0;
        double y2finalmove=0;
        double xtoright=0;
        double ytoright=0;
        double xborder=0;
        double xwinningmove=0;
        double ywinningmove=0;
        double currentxstart=0;
        double currentystart=0;
        //double[] centroid1;
        //      Point Centroid1, Centroid2;
        int angle2final = 0;
        double xtostartwith = 0;
        //if (shape1.isConvex())
        //    return null;
        double d=1E-6;
        Polygon LargeShape1=shape1.dilate(d);
        for ( int i=0; i<360; i=i+5 )
            {
                double angleToMove = Math.toRadians( i );
                Polygon shapeRotated2 = shape2.rotate( angleToMove );
 
                for ( int j=0; j<LargeShape1.n(); j++) {
                    double angle1 = LargeShape1.angle(j);
                    for ( int k=0; k<shapeRotated2.n(); k++) {
                        double angle2 = shapeRotated2.angle(k);
                        if (angle1 + angle2 < 360) {
                            double x1 = LargeShape1.x(j);
                            double y1 = LargeShape1.y(j);
                            double x2 = shapeRotated2.x(k);
                            double y2 = shapeRotated2.y(k);
                            xtomove = x1-x2;
                            ytomove = y1-y2;
                            Polygon shapeTranslated2 =
                                shapeRotated2.translate(xtomove, ytomove);
                            if (!(shape1.intersects(shapeTranslated2))) {
                                //System.out.println( "A possible answer!");
                                // I should really create a function that finds the min x with y>1
 
                                //double minx1; double miny1; double maxx1; double maxy1;
                                // double minx2; double miny2; double maxx2; double maxy2;
                                //double minx; double miny; double maxx; double maxy;
                                Rectangle bd1, bd2, bd;
                                double xdist; double ydist;
                                // minx1=shape1.minx(); miny1=shape1.miny(); maxx1=shape1.maxx(); maxy1=shape1.maxy();
                                // minx2=shapeTranslated2.minx(); miny2=shapeTranslated2.miny();
                                // maxx2=shapeTranslated2.maxx(); maxy2=shapeTranslated2.maxy();
                                bd1 = shape1.bound();
                                bd2 = shapeTranslated2.bound();
 
                                // if (minx1<minx2) minx=minx1; else minx=minx2;
                                // if (miny1<miny2) miny=miny1; else miny=miny2;
                                // if (maxx1>maxx2) maxx=maxx1; else maxx=maxx2;
                                // if (maxy1>maxy2) maxy=maxy1; else maxy=maxy2;
                                bd = bd1.union( bd2 );
                                // xdist=maxx-minx; ydist=maxy-miny;
                                xdist = bd.width(); ydist = bd.height();
                                //System.out.println( "ydist"+ydist);
                                if (!(ydist > 1 )) {
                                    // System.out.println( "currentmin"+currentminx+"xdist"+xdist);
                                    if (((testnum==1) || (xdist*ydist < currentminx)) && (xdist > 0)) {
                                        testnum=2;
                                        currentminx = xdist*ydist;
                                        xwinningmove=xdist+d;
                                        ywinningmove=ydist+d;
                                        xborder = Math.abs(xdist + d);
                                        xtoright = -bd.x0(); // -minx;
                                        ytoright = -bd.y0(); // -miny;
                                        x2finalmove = xtomove;
                                        y2finalmove = ytomove;
                                        angle2final = i;
                                        //Centroid1 = shape1.centroid();
                                        //Centroid2 = shapeTranslated2.centroid();
                                        // System.out.println("CurrentMinX: "+currentminx+" angle "+angle2final);
                                        //      System.out.println( "I am here.!!!!!!!!");
                                    }
 
                                }
                            }
                        }
                    }
                }
            }
        if (currentminx > 0) {
            // System.out.println( "Giving Solution");
            double botxstart = 0;
            Solution ret = new Solution( _nc );
            for ( int j=0; j<_nc; j++ )
                {
                    /*              if (j%2 == 1) {
                        ret.x[j] = xtoright + xborder*(((j+1)/2)-1);//+xtostartwith;// + Centroid1.x();
                        ret.y[j] = ytoright;//Centroid1.y();
                        ret.angle[j] = 0;
                    }
                    else {
                        ret.x[j] = x2finalmove+xtoright + xborder*((j/2)) ; //+xtostartwith;// + Centroid2.x();
                        ret.y[j] = y2finalmove+ytoright;//Centroid2.y();
                        ret.angle[j] = Math.toRadians( angle2final );
                        xtostartwith = xtostartwith + currentminx + d;
                    }
                    */
 
                    if (((j%2 == 0)&&(extrapiece==0)) || ((j%2==1) && (extrapiece==1))) {
                        ret.x[j] = currentxstart+xtoright+DELTA;
                        ret.y[j] = currentystart + ytoright;//Centroid1.y();
                        ret.angle[j] = 0;
 
                    }
                    else {
                        ret.x[j] = currentxstart+ x2finalmove + xtoright+DELTA;
                        ret.y[j] = currentystart + y2finalmove + ytoright;
                        ret.angle[j] = Math.toRadians( angle2final );
                        currentystart=currentystart+ywinningmove;
                        if (((currentystart+ywinningmove)> 1+d) && (j<_nc-1)) {
                            //Check to see if you can put another piece in
                            // and change the
                            double doughleftover = 1-currentystart;
                            if ((xwinningmove < doughleftover)
                                && (botxstart + ywinningmove <
                                    currentxstart+xwinningmove)){
                                System.out.println("xwinningmove="+xwinningmove+" dough="+doughleftover);
                                j++;
 
                                ret.x[j] = botxstart+y2finalmove+ytoright+DELTA;
                                    //   x2finalmove + xtoright;
                                ret.y[j] = currentystart+xtoright;// + x2finalmove+xtoright; // y2finalmove;// +  ytoright;
                                ret.angle[j] = Math.toRadians( 90 );
                                if (j<_nc-1) {
                                    j++;
                                    ret.x[j] = botxstart + ytoright+DELTA;//+ x2finalmove + xtoright;
                                    ret.y[j] = currentystart + xtoright + x2finalmove;// + y2finalmove + ytoright;
                                    ret.angle[j] = Math.toRadians( angle2final +90);
                                    botxstart = botxstart+ywinningmove;
                                }
                            }
                            currentystart=0;
                            currentxstart=currentxstart+xwinningmove;
                        }
                        xtostartwith = xtostartwith + currentminx + d;
                    }
                }
            //xtostartwith=xtostartwith+0.1;
            ret.rbound=currentxstart+xwinningmove+DELTA+DELTA+DELTA;
            return ret;
        }
                     
        else return null;
    }
}

/** A simplified region class used in the Cookie Cutter Problem
 */
class SimpleRegion {
    Polygon pg;
    Polygon[] holes;    
    Rectangle bound;

    SimpleRegion( Polygon p, Polygon[] h ) {
        pg = p;
        holes = h;
        bound = p.bound();
    }

    public String toString() {
        StringBuffer sb = new StringBuffer();
        
        sb.append( pg.toString() );
        if ( holes != null )
        {
            for ( int i=0; i<holes.length; i++ )
            {
                sb.append( holes[i].toString() );
                sb.append( "-" );
            }
        }
        
        return sb.toString();
    }

    static String toString( SimpleRegion[] regs ) {
        if ( regs == null )
            return "{null}";

        StringBuffer sb = new StringBuffer();
        sb.append( "{ " );
        for ( int i=0; ; i++ )
        {
            sb.append( regs[i].toString() );
            if ( i >= regs.length - 1 )
                break;
            sb.append( ", " );
        }
        sb.append( " }" );
        return sb.toString();
    }
}

