/**
 * @author
 *  Behrooz Badii 	bb2122@columbia.edu
 *  Hanhua Feng         hanhua@cs.columbia.edu
 *  Edan Harel 	        edan@columbia.edu
 */
package CookieCutter.g6;

import CookieCutter.*;
import CookieCutter.g6.*;

/** The wrapper program of the 2d geometry classes.  It implements
 * the game player.
 */
public abstract class Group6Wrapper implements IFCPlayer {
    public static double DELTA = CookieCutterProblem.DELTA; // the gap

    CookieCutter _game;                   // the mediator class
    Vertex[][]   _cookies;                // original cookie shapes
    int          _copies;                 // number of cookies
    String       _name = "The Lord of the Cookie Cutters";

    /** register the game and get game parameters 
     */
    public void register( CookieCutter game ) throws Exception {
        _game = game;
        _cookies = game.cookieshapes();
        _copies = game.cookieCopies();
    }

    /** the name of the player, which might be changed by derived classes
     */
    public String name() throws Exception {
        return _name;
    }

    /** the game is played by the program
     */
    public boolean interactive() throws Exception {
        return false;
    }

    /** check the validity of a solution, use system intersectio functions
     */ 
    public boolean valid( Polygon pg, CookieCutterProblem.Solution soln ) {
        if ( null == soln )
            return false;

        try
        {
            int n = pg.n();
            for ( int i=0; i<_copies; i++ )
            {
                Polygon pg0 = pg.place( soln.x[i], soln.y[i], soln.angle[i] );

                Vertex[] vt0 = new Vertex[n];
                for ( int k=0; k<n; k++ )
                {
                    vt0[k] = new Vertex( pg0.x(k), pg0.y(k) );
                    if ( pg0.x(k) <= 0 || pg0.x(k) >= soln.rbound
                         || pg0.y(k) <= 0 || pg0.y(k) >= 1 )
                        return false;
                }
                
                for ( int j=0; j<i; j++ )
                {
                    Polygon pg1 = pg.place( soln.x[j], soln.y[j], soln.angle[j] );
                    Vertex[] vt1 = new Vertex[n];
                    for ( int k=0; k<n; k++ )
                        vt1[k] = new Vertex( pg1.x(k), pg1.y(k) );
                    
                    for ( int k=0, pk=n-1; k<n; pk=k++ )
                        for ( int l=0, pl=n-1; l<n; pl=l++ )
                            if ( _game.intersects( vt0[pl], vt0[l], 
                                                   vt1[pk], vt1[k] ) )
                                return false;
                    
                }
            }
        }
        catch ( Exception e )
        {
            return false;
        }

        return true;
    }

    /** the abstract method of solver that the derived class must override
     */
    abstract public CookieCutterProblem.Solution solve( 
        CookieCutterProblem prob, Polygon pg );

    /** generate a Polygon object
     */
    public static Polygon createPolygon( Vertex[] vet ) throws Exception {
        Point[] pts = new Point[ vet.length ];
        for ( int i=0; i<pts.length; i++ )
            pts[i] = new Point( vet[i].xpos(), vet[i].ypos() );
        return new Polygon( pts );
    }

    /** the moves generator.
     */
    public Move[] moves() throws Exception {
        Move[] ret = new Move[ _cookies.length ];

        // create the problem
        CookieCutterProblem prob = new CookieCutterProblem( 1.0, _copies );

        // for each shape call the problem solver
        for ( int i=0; i<_cookies.length; i++ )
        {
            // find the centroid, and re-center the polygon to the centroid,
            // the dilate the polygon by DELTA for a small gap
            Vertex centroid = _game.centroid( _cookies[i] );
            Polygon pg = createPolygon( _cookies[i] ).translate( 
                -centroid.xpos(), -centroid.ypos() );

            // call solver to solve this problem
            CookieCutterProblem.Solution soln;
            try 
            {
                // the solver should call valid() function here to
                // verify the result, and dilate itself
                soln = solve( prob, pg );
            } 
            catch( Exception e ) 
            {
                e.printStackTrace();
                soln = null;
            }

            ret[i] = new Move(_copies);
            if ( null == soln )
            {
                // the solver can not find a solution, what can we do?
                System.err.println( 
                    "Fatal error: unable to place cookie #" + i );
                ret[i].setRightBoundary( 0 );
                continue;
            }

            // we now have a solution.  Generate the moves
            for ( int j=0; j<_copies; j++ )
            {
                ret[i].setCookiePosition( 
                    j, new Vertex( soln.x[j], soln.y[j] ), soln.angle[j] );
            }            
            ret[i].setRightBoundary( soln.rbound );
        }

        return ret;
    }
}
