/**
 * @author
 *  Behrooz Badii 	bb2122@columbia.edu
 *  Hanhua Feng         hanhua@cs.columbia.edu
 *  Edan Harel 	        edan@columbia.edu
 */
package CookieCutter.g6;

import java.io.Serializable;

/** 
 * An infinite line on the xy-plane
 * 
 * Note: A line has a direction!  Here the line equation ax+by=c is
 * used.  If b is greater than zero, points at the up-side of this
 * line satisfy ax+by>c.  If b is less than zero, the down-side of the
 * line satisfy ax+by>c.  If b is zero, in which case the line is
 * vertical, then the left side of the line satisfy ax+by>c when a>0.
 * Two parameters a and b cannot be both zero.
 */
class Line implements Serializable {
    public static Line X_AXIS = new Line( 0, 1, 0 );
    public static Line Y_AXIS = new Line( 1, 0, 0 );

    protected double _a, _b, _c;   // The line of ax+by=c such that a*a+b*b=1

    /** create a line by the equation ax+by=c
     */
    public Line( double a, double b, double c ) {
        double d = Math.sqrt( a*a + b*b );
        if ( d <= Plane.EPS )
        {
            _a = 0;
            _b = 1;
            _c = c;
        }
        else
        {
            d = 1.0 / d;

            _a = a * d;
            _b = b * d;
            _c = c * d;
        }
    }

    /** given two points, create a line.
     *
     * Note: we keep let the sign of b be the same as x_1-x_0, so that
     * the left-side points of the vector (p0->p1) always satisfy 
     * ax+by>c.  It also works for x_1-x_0 = 0.
     */
    public Line( Point p0, Point p1 ) {
        this( p0.y() - p1.y(),
              p1.x() - p0.x(),
              p0.y() * p1.x() - p1.y() * p0.x() );
    }

    public final double a() {
        return _a;
    }

    public final double b() {
        return _b;
    }

    public final double c() {
        return _c;
    }

    /** the distance to a point
     */
    public double distance( Point p ) {
        return Math.abs( _a * p.x() + _b * p.y() - _c );
    }

    /** shift a line to the left-hand side by delta
     */
    public Line shift( double delta ) {
        return new Line( _a, _b, _c + delta );
    }

    /** get the foot of p to this line.
     */
    public Point foot( Point p ) {
        if ( Math.abs( _a*p.x() + _b*p.y() - _c ) <= Plane.EPS )
            return p;
        
        return (Point) crosspoint( new Line( _b, -_a, _b*p.x() - _a*p.y() ) );
    }

    /** return the cross point of this line and line ln.
     *  It returns null if the two lines is Parellel, and a line
     *  if two lines are the same.
     */
    public final Object crosspoint( Line ln ) {
        double det = _a * ln._b - _b * ln._a;

        if ( Math.abs(det) <= Plane.EPS * 10 ) 
        {
            if ( Math.abs( _c - ln._c ) > Plane.EPS 
                 || _a * ln._a + _b * ln._b <= 0 )
                return null;
/*            
            System.out.println( "Same lines: "
                                + new Line(_a,_b,_c).toString() 
                                + " vs "
                                + new Line(ln._a,ln._b,ln._c).toString() );
*/
            return this;
        }

        det = 1.0 / det;        
        return new Point( ( ln._b * _c - _b * ln._c ) * det,
                          ( _a * ln._c - ln._a * _c ) * det );
    }

    /** the angle of two lines.  Note that lines are directed,
     * @return the angle from this line to line ln, in a range of
     * (-\pi,\pi)
     */
    double angle( Line ln ) {
        double a = Math.atan2( _a, _b ) - Math.atan2( ln._a, ln._b );
        if ( a > Math.PI )
            return a - Math.PI * 2;
        if ( a <= -Math.PI )
            return a + Math.PI * 2;
        return a;
    }

    /** check the point p is on which side of the directed segment
     * (v0->v1).  It returns 1 if this point is on the left-hand side,
     * and -1 if it is on the right-hand side.  If the point is almost
     * on the infinite line defined p0 and p1, this function returns
     * 0.
     */
    public int whichSide( Point p ) {
        return Plane.sign( _a * p.x() + _b * p.y() - _c );
    }

    /** move the point p along the direction of this line by d
     */
    public Point translateAlong( Point p, double d ) {
        return new Point( p.x()-_b*d, p.y()+_a*d );
    }

    public String toString() {
        StringBuffer sb = new StringBuffer();
        
        sb.append( "{(" );
        sb.append( Plane.format(_a) );
        sb.append( ")x+(" );
        sb.append( Plane.format(_b) );
        sb.append( ")y=" );
        sb.append( Plane.format(_c) );
        sb.append( "}" );

        return new String(sb);
    }
}
